/*
Module : ATLCPImplMT.h
Purpose: Implementation for a freeware replacement class for the ATL class IConnectionPointImpl which performs 
         automatic marshaling of client interfaces across COM apartments using the GIT (Global Interface
         Table). Put more simply, this class will allow you to fire events from your COM objects from
         a different thread than the one on which the object was initially created. The code was based
         on the MS Article "SAMPLE: ATLCPImplMT encapsulates ATL event firing across COM apartments" at
         http://support.microsoft.com/kb/280512. Based on this original code, I have performed an 
         extensive rewrite of the class using the ATL 8 "CComGITPtr" class. In addition it has been updated 
         based upon the ATL 8 "IConnectionPointImpl" implementation. It also uses ATL::CSimpleArray
         to hold the GIT'ed interfaces rather than the CComDynamicUnkArray class.

         To use the class in your ATL client code, you first need to perform the usual sequence of steps
         to support events in your ATL project. This will result in the creation of a proxy class which is
         derived from IConnectionPointImpl which contains a number of "Fire_{EventName}" methods which you 
         would then derive from in your normal ATL objects, thus allowing you to fire events back to your 
         clients. You should change the inheritance of this proxy class to be derived from IConnectionPointImplMT 
         instead of IConnectionPointImpl. Note that the template parameters for IConnectionPointImplMT are 
         slightly different than IConnectionPointImpl. The template parameters for IConnectionPointImplMT 
         are as follows:

         template <class T, const IID* piid, class TSinkInterface> class IConnectionPointImplMT.

         Also for each generated function (named Fire_{EventName}) within the proxy class, you will need to 
         rework the firing code similar to the following:

         HRESULT Fire_OnRMCSentence()
         {
           CComCritSecLock<CComAutoCriticalSection> lock(m_CPMTCritSec);

           CComVariant varResult;
           int nConnections = m_Clients.GetSize();
           for (int nConnectionIndex=0; nConnectionIndex<nConnections; nConnectionIndex++)
           {
             ATL::CComPtr<IYourEventsInterface> sp;
             const HRESULT hr = this->GetInterfaceAt(nConnectionIndex, sp);
             if (SUCCEEDED(hr))
             {
               Code to call sp->Invoke with correct parameters
               ......
               ......
               sp->Invoke(.....);
             }
           }
           return .....;
         }


         That's all you need to do. All other client remains the same. For a concrete example of using the 
         class in client code, please see the GPSComCP.h file in the GPSCom download. Happy COM coding!

Created: PJN / 21-06-2002
History: PJN / 20-07-2006 1. Updated code to clean compile on VC 2005
         PJN / 20-01-2007 1. Major rewrite of the code as documented above in the Purpose section
         PJN / 08-02-2007 1. Updated copyright details.
                          2. How you implement the Fire_ methods as documented above has been updated to refer
                          to the need to serialize access to the "m_Clients" member variable which is 
                          accessed via the GetInterfaceAt method. The CProxy_IGPSConnectionEvents class in the 
                          GPSCom project and the IConnectionPointImplMT::GetInterfaceAt method have been 
                          updated to address this issue. Thanks to Michael Ford for reporting this bug. 
                          3. The actual pointers which are stored on the internal m_Clients array are now
                          "CComGITPtr<TSinkInterface>" heap pointers rather than actual stack instances. By 
                          doing this, the code now has explicit control over when and how copy constructors and 
                          operator= methods are being called as we add a element to the array. This avoids extra
                          overhead as the interface pointer is being added to the GIT. It also addresses an issue
                          where the cookie value which was being returned by Advise was actually incorrect due to
                          the fact that a copy of the CComGITPtr was being added to the array. Thanks to
                          Michael Ford for reporting this issue.
         PJN / 21-03-2010 1. Updated copyright details.
                          2. Updated the instructions on how to code the Fire_* methods to avoid a COM reference 
                          counting problem which could have resulted in memory leaks. Thanks to Peter Blanz for
                          reporting this issue.
         PJN / 25-04-2010 1. Updated to work when "using namespace ATL" is now explicitly set in client code. 
                          2. GetInterfaceAt method has been reworked to surface a HRESULT as a return value. In 
                          addition the IDispatch pointer is now returned already wrapped in a CComPtr<IDispatch> 
                          smart pointer reference parameter. The instructions of how you recode the Fire_* methods 
                          have also been reworked above.
         PJN / 17-12-2015 Reworked the IConnectionPointImplMT::GetInterfaceAt method to have the following signature: 
                          HRESULT GetInterfaceAt(int nConnectionIndex, CComPtr<TSinkInterface>& pTSinkInterface)
                          This change from returning a CComPtr<IDispatch>& means that the code can handle client
                          situations where an IDispatch interface is not available. This can occur where the 
                          event interface is implemented as a marshalled C# class. Thanks to David Korn for 
                          reporting this issue.
         PJN / 18-11-2016 1. Updated copyright details
         PJN / 07-03-2017 1. Updated copyright details
         PJN / 18-05-2017 1. Updated the code to compile cleanly using /permissive-.
         PJN / 13-10-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code will
                          now only compile on VC 2017 or later.
         PJN / 17-05-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
                          3. Reworked the IConnectionPointImplMT::m_Clients array to store the store the cookie instead of 
                          the GIT pointers since you can always get the GIT pointers from them. This also allowed easier 
                          lookups so a simpler implementation of Unadvise can be made. Thanks to Michael Ford for this nice
                          addition.
                          4. Fixed an erroneous semicolon in the sample code for Fire_OnRMCSentence in the comments in this
                          module. Thanks to Michael Ford for reporting this issue.
         PJN / 04-04-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 01-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2002 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __CPIMPLMT_H__
#define __CPIMPLMT_H__


//////////////////// Includes /////////////////////////////////////////////////

#ifndef __ATLCOM_H__
#pragma message("To avoid this message, please put atlcom.h in your pre compiled header (normally stdafx.h)")
#include <atlcom.h>
#endif //#ifndef __ATLCOM_H__


//////////////////// Classes //////////////////////////////////////////////////

namespace ATL
{

template <class T, const IID* piid, class TSinkInterface, class TASinkInterface = IUnknown>
class ATL_NO_VTABLE IConnectionPointImplMT : public _ICPLocator<piid>
{
  using CComEnumConnections = CComEnum<IEnumConnections, &__uuidof(IEnumConnections), CONNECTDATA, _Copy<CONNECTDATA>>;

public:
//Constructors / Destructors
#pragma warning(suppress: 26455)
  IConnectionPointImplMT() = default;
  IConnectionPointImplMT(const IConnectionPointImplMT&) = delete;
  IConnectionPointImplMT(IConnectionPointImplMT&&) = delete;
#pragma warning(suppress: 26432)
  ~IConnectionPointImplMT()
  {
    //Serialize access to the array
#pragma warning(suppress: 26447)
    const CComCritSecLock<CComAutoCriticalSection> lock{m_CPMTCritSec};

    //Tidy up the client array
#pragma warning(suppress: 26447)
    const int nClients{m_Clients.GetSize()};
    for (int i{nClients-1}; i>=0; i--)
    {
#pragma warning(suppress: 26447)
      CComGITPtr<TSinkInterface> pClient(m_Clients[i]); //Attach and Revoke
    }
  }

//Methods
  IConnectionPointImplMT& operator=(const IConnectionPointImplMT&) = delete;
  IConnectionPointImplMT& operator=(IConnectionPointImplMT&&) = delete;

#pragma warning(suppress: 26433 26487)
  STDMETHOD(_LocCPQueryInterface)(_In_ REFIID riid, _COM_Outptr_ void** ppvObject)
  {
  #ifndef _ATL_OLEDB_CONFORMANCE_TESTS
#pragma warning(suppress: 26477)
    ATLASSERT(ppvObject != nullptr);
  #endif //#ifndef _ATL_OLEDB_CONFORMANCE_TESTS
    if (ppvObject == nullptr)
      return E_POINTER;
    *ppvObject = nullptr;

    if (InlineIsEqualGUID(riid, __uuidof(IConnectionPoint)) || InlineIsEqualUnknown(riid))
    {
      *ppvObject = this;
      this->AddRef();
    #ifdef _ATL_DEBUG_INTERFACES
      _AtlDebugInterfacesModule.AddThunk(reinterpret_cast<IUnknown**>(ppvObject), _T("IConnectionPointImplMT"), riid);
    #endif //#ifdef _ATL_DEBUG_INTERFACES
      return S_OK;
    }
    else
      return E_NOINTERFACE;
  }

#pragma warning(suppress: 26440)
  STDMETHOD(GetConnectionInterface)(_Out_ IID* piid2) noexcept
  {
    if (piid2 == nullptr)
      return E_POINTER;
    *piid2 = *piid;
    return S_OK;
  }

  STDMETHOD(GetConnectionPointContainer)(_Outptr_ IConnectionPointContainer** ppCPC) noexcept
  {
#pragma warning(suppress: 26429 26466)
    T* pT{static_cast<T*>(this)};

    //No need to check ppCPC for null since QI will do that for us
#pragma warning(suppress: 26490)
    return pT->QueryInterface(IID_IConnectionPointContainer, reinterpret_cast<void**>(ppCPC));
  }

  STDMETHOD(Advise)(_Inout_ IUnknown* pUnkSink, _Out_ DWORD* pdwCookie)
  {
    //Validate our parameter
    if ((pUnkSink == nullptr) || (pdwCookie == nullptr))
      return E_POINTER;

    //Default the output parameters
    *pdwCookie = 0;

    //Check that the client implements the expected sink interface
    CComPtr<TSinkInterface> pClientSink;
    HRESULT hr{pUnkSink->QueryInterface(&pClientSink)};
    if (SUCCEEDED(hr))
    {
      //There shouldn't be a need to store the pointers, we just need the cookie
      //The following registers the sink in the GIT for you
      //NOTE: if you use the constructor passing in the interface, a CAtlException can be thrown
      CComGITPtr<TSinkInterface> pNewClient;
      hr = pNewClient.Attach(pClientSink); //Attach instead, no exception
      if (SUCCEEDED(hr))
      {
        //And then get the corresponding cookie for the specified interface
        const DWORD dwCookie{pNewClient.GetCookie()};

        //Serialize access to the array as we are about to add a new client
        const CComCritSecLock<CComAutoCriticalSection> lock{m_CPMTCritSec};

        //Finally add to the array of clients
        if (!m_Clients.Add(dwCookie))
        {
          hr = E_FAIL;
          //destructor will revoke the interface
        }
        else
        {
          pNewClient.Detach(); //we don't want to revoke!
          *pdwCookie = dwCookie;
        }
      }
    }
    else if (hr == E_NOINTERFACE)
      hr = CONNECT_E_CANNOTCONNECT;

    return hr;
  }

  STDMETHOD(Unadvise)(_In_ DWORD dwCookie)
  {
    //Serialize access to the array
    const CComCritSecLock<CComAutoCriticalSection> lock{m_CPMTCritSec};

    //Remove the specified client from the array of clients
    if (!m_Clients.Remove(dwCookie))
      return CONNECT_E_NOCONNECTION;

    //Revoke the interface
    CComGITPtr<TSinkInterface> pClient{dwCookie};

    return S_OK;
  }

#pragma warning(suppress: 26436 26487)
  STDMETHOD(EnumConnections)(_COM_Outptr_ IEnumConnections** ppEnum)
  {
    //Validate our parameters
    if (ppEnum == nullptr)
      return E_POINTER;

    //Set the output parameter to a default value
    *ppEnum = nullptr;

    //Create the COM object which will handle the COM interface being returned
    CComObject<CComEnumConnections>* pEnum{nullptr};
#pragma warning(suppress: 26400 26409)
    ATLTRY(pEnum = new CComObject<CComEnumConnections>)
    if (pEnum == nullptr)
      return E_OUTOFMEMORY;

    //Serialize access to the clients array
    CComCritSecLock<CComAutoCriticalSection> lock{m_CPMTCritSec};

    //Allocate the array
    const int nClients{m_Clients.GetSize()};
    CONNECTDATA* pcd{nullptr};
#pragma warning(suppress: 26400 26409)
    ATLTRY(pcd = new CONNECTDATA[nClients])
    if (pcd == nullptr)
    {
      delete pEnum;
      return E_OUTOFMEMORY;
    }

    //Copy the valid CONNECTDATA's
    CONNECTDATA* pend{pcd};
    HRESULT hr{S_OK};
    for (int i{0}; i<nClients; i++)
    {
      const DWORD dwCookie{m_Clients[i]};
      CComGITPtr<TSinkInterface> pClient{dwCookie};
      CComPtr<TSinkInterface> sinkInterface;
      hr = pClient.CopyTo(&sinkInterface);
      pClient.Detach(); //We don't want to revoke!
      if (SUCCEEDED(hr))
      {
#pragma warning(suppress: 26486 26489)
        hr = sinkInterface.QueryInterface(&pend->pUnk);
        if (SUCCEEDED(hr))
        {
          //Also set the cookie value and move on to the next element in the output CONNECTDATA array
#pragma warning(suppress: 26489)
          pend->dwCookie = dwCookie;
#pragma warning(suppress: 26481)
          pend++;
        }
      }
    }
    //don't copy the data, but transfer ownership to it
#pragma warning(suppress: 26486)
    hr = pEnum->Init(pcd, pend, nullptr, AtlFlagTakeOwnership);
    lock.Unlock();
    if (FAILED(hr))
      delete pEnum;
    else
    {
#pragma warning(suppress: 26490)
      hr = pEnum->_InternalQueryInterface(IID_IEnumConnections, reinterpret_cast<void**>(ppEnum));
      if (FAILED(hr))
        delete pEnum;
    }
    return hr;
  }

  HRESULT GetInterfaceAt(_In_ int nConnectionIndex, _Out_ CComPtr<TSinkInterface>& pTSinkInterface)
  {
    //It is assumed that client code has already taken a lock on m_CPMTCritSec
    //at this point to ensure we can safely access m_Clients directly
    const DWORD dwCookie{m_Clients[nConnectionIndex]};
    //Delegate to "GetInterfaceFor"
    return GetInterfaceFor(dwCookie, pTSinkInterface);
  }

  HRESULT GetInterfaceFor(_In_ DWORD dwCookie, _Out_ CComPtr<TSinkInterface>& pTSinkInterface) noexcept
  {
    CComGITPtr<TSinkInterface> pClient{dwCookie};
    const HRESULT hr{pClient.CopyTo(&pTSinkInterface)};
    pClient.Detach(); //we don't want this revoked!
    return hr;
  }

  HRESULT GetAsyncInterfaceFor(_In_ DWORD dwCookie, _Out_ CComPtr<TASinkInterface>& pTASinkInterface)
  {
    CComGITPtr<TSinkInterface> pClient{dwCookie};
    if (pClient)
    {
      CComPtr<TSinkInterface> sp;
      pClient.CopyTo(&sp);
      pClient.Detach(); //we don't want this revoked!

      CComQIPtr<ICallFactory> pCF{sp};
      if (pCF)
      {
        CComPtr<IUnknown> pUnk;
        if (SUCCEEDED(pCF->CreateCall(__uuidof(TASinkInterface), nullptr, IID_ISynchronize, &pUnk)))
          return pUnk.QueryInterface(&pTASinkInterface);
      }
    }
    return E_NOINTERFACE;
  }

//Member variables
  CSimpleArray<DWORD> m_Clients; //The array of GIT cookies we are managing
  CComGlobalsThreadModel::AutoCriticalSection m_CPMTCritSec; //The critical section to serialize access to "m_Clients" above
};


} // namespace ATL


#endif //#ifndef __CPIMPLMT_H__
