/*
Module : GPSCom2Client.h
Purpose: C++ client wrapper class for the GPSCom2 OLE Server. This provides comprehensive support for the 
         following NMEA sentences: $GPRMC, $GPGGA, $GPGSV, $GPGLL, $GPVTG, $GPGSA, $GPGRS & $GPGST
Created: PJN / 27-04-2010
History: PJN / 01-05-2010 1. Updated the CSentence::Time method to correctly handle centiseconds values
                          2. Updated the CSentence::Angle method to correctly handle empty fields  
                          3. Fixed logic errors in CVTGSentence::Parse
                          4. CSentence::Tokenize now removes all fields from the array before it starts
                          parsing
                          5. All places which parse a floating point number now handle empty fields
                          6. Fixed a logic error in CSentence::Angle
         PJN / 17-12-2015 1. Updated copyright details.
                          2. Updated sample app to clean compile in VC 2010 - 2015.
                          3. Reworked the classes to optionally compile without MFC. By default the class now
                          use STL classes and idioms but if you define CGPSCOM2CLIENT_MFC_EXTENSIONS the class
                          will revert back to the MFC behaviour.
         PJN / 19-11-2016 1. Updated copyright details
                          2. Added support for the GPSPort::CreateConnection2 method.
                          3. Added SAL annotations to the code
                          4. Fixed issue in the message when the code detects you do not have atlcom.h included.
                          Thanks to Frits van Veen for reporting this issue.
         PJN / 07-03-2017 1. Updated copyright details.
                          2. Updated code to clean compile in VC 2017
                          3. Updated comment on the #import "libid:.." line to tell the user how to resolve the
                          issue.
         PJN / 07-11-2017 1. Replaced CString::operator LPC*STR() calls with CString::GetString calls.
                          2. Added missing targetver.h file to the zip download file.
                          3. Included information in GPSCom2Client.h about registering GPSCom2.
                          4. Fixed a compile issue in GPSCom2Client.h when _ATL_NO_AUTOMATIC_NAMESPACE is 
                          defined.
                          5. Added validation code to each of the CSentence::Parse methods to verify the size 
                          of the "sFields" array. Thanks to Frits van Veen for reporting this bug.
         PJN / 13-10-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code will
                          now only compile on VC 2017 or later.
                          3. Replaced BOOL with bool throughout the codebase
                          4. Removed code which supported CGPSCOM2CLIENT_MFC_EXTENSIONS code path
         PJN / 23-06-2019 1. Fixed compile warnings in GPSCom2Client.h when importing GPSCom2 type library.
                          2. Replaced enum with enum class throughout the codebase
         PJN / 30-06-2019 1. Added support for GNSS, GLONASS, Galileo and BeiDou variants of NMEA sentences
                          in addition to the standard GPS versions.
                          2. Added support for G*GNS & G*ZDA NMEA sentences.
         PJN / 09-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 04-11-2019 1. Updated initialization of various structs to use C++ 11 list initialization
         PJN / 04-04-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 01-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.
         PJN / 17-06-2023 1. Updated copyright details.
                          2. Made a local variable in GPSCom2::CSentence::IsChecksumValid const
                          3. Removed unnecessary constructors from sentence classes.

Copyright (c) 2010 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __GPSCOM2CLIENT_H__
#define __GPSCOM2CLIENT_H__

#ifndef CGPSCOM2CLIENT_EXT_CLASS
#define CGPSCOM2CLIENT_EXT_CLASS
#endif


//////////////////// Includes /////////////////////////////////////////////////

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_
#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

//Pull in the definitions for GPSCom2
#pragma warning(push)
#pragma warning(disable: 4471 26496 26477 26486)
#import "libid:2DC6E5DC-6B8B-4E82-ABE5-711926DFA19B" rename_namespace("GPSCom2"), raw_interfaces_only //If you get a compilation error from this line, then you need to register GPSCom2.exe for the bitness you are compiling for. Please see gpscom2.htm for information on how to register GPSCom2.
#pragma warning(pop)

#ifndef __ATLCOM_H__
#pragma message("To avoid this message, please put atlcom.h in your pre compiled header (normally stdafx.h)")
#include <atlcom.h>
#endif //#ifndef __ATLCOM_H__


//////////////////// Implementation ///////////////////////////////////////////

namespace GPSCom2 //Lets put the client classes / enums structs etc into the same namespace as where we put the #import classes into
{
//Typedefs
#ifdef _UNICODE
  using String = std::wstring;
#else
  using String = std::string;
#endif //#ifdef _UNICODE
  using StringArray = std::vector<String>;

//structure required by SINK_ENTRY_INFO to implement event sinking using ATL's IDispEventSimpleImpl class
static ATL::_ATL_FUNC_INFO Event1Info{CC_STDCALL, VT_EMPTY, 1, { VT_BSTR }};

enum class BOOLEAN
{
  NMEA_UNKNOWN,
  NMEA_TRUE,
  NMEA_FALSE
};

enum class EASTWEST
{
  NMEA_EW_UNKNOWN = 0,
  NMEA_EAST,
  NMEA_WEST
};

enum class NORTHSOUTH
{
  NMEA_NS_UNKNOWN = 0,
  NMEA_NORTH,
  NMEA_SOUTH
};

enum class FIX
{
  NMEA_FIX_NOT_AVAILABLE = 0,
  NMEA_GPS_FIX = 1,
  NMEA_DIFFERENTIAL_FIX = 2
};

enum class SELECTION_MODE
{
  NMEA_SELECTION_MODE_MANUAL = 1,
  NMEA_SELECTION_MODE_AUTOMATIC = 2,
  NMEA_SELECTION_MODE_UNKNOWN = 3  
};

enum class MODE
{
  NMEA_MODE_UNKNOWN = 0,
  NMEA_MODE_NOFIX = 1,
  NMEA_MODE_2DFIX = 2,
  NMEA_MODE_3DFIX = 3,
};

enum class RESIDUAL_USE
{
  NMEA_RESIDUAL_USE_UNKNOWN = -1,
  NMEA_RESIDUAL_USE_IN_GGA = 0,
  NMEA_RESIDUAL_USE_AFTER_GGA = 1,
};

struct TIME
{
  int nHour;
  int nMinute;
  int nSecond;
  int nCentiSeconds;
};

struct DATE
{
  int nYear;
  int nMonth;
  int nDay;
};

struct ANGLE
{
  WORD wDegrees;
  WORD wMinutes;
  WORD wSeconds;
  WORD wCentiSeconds;
  double dValue; //A simple numeric version of the angle in degrees
  bool bValid;
};

struct LATITUDE
{
  ANGLE Value;
  NORTHSOUTH Northing;
  double dValue; //A simple numeric version of the latitude (positive values are north of the equator)
  bool bValid;
};

struct LONGITUDE
{
  ANGLE Value;
  EASTWEST Easting;
  double dValue; //A simple numeric version of the latitude (positive values are east of the Greenwich prime meridian)
  bool bValid;
};


//////////////////// Classes //////////////////////////////////////////////////

//A helper class for parsing NMEA sentences
class CSentence
{
public:
//Methods
  static void Tokenize(_In_ const String& sSentence, _Inout_ StringArray& sFields)
  {
    //Remove all the fields before we go any further
    sFields.clear();

    int nCurrentPos{0};
    while (nCurrentPos != -1)
    {
      //Work out the position of the next delimiter
      auto nEndPos{sSentence.find(_T(','), nCurrentPos)};
      if (nEndPos == String::npos)
        nEndPos = sSentence.find(_T('*'), nCurrentPos);

      //Pull out the field
      String sField;
      if (nEndPos != String::npos)
        sField = sSentence.substr(nCurrentPos, nEndPos - nCurrentPos);
      else
        sField = sSentence.substr(nCurrentPos, sSentence.length() - nCurrentPos);

#pragma warning(suppress: 26489)
      sFields.push_back(sField);

      //Prepare for the next time around
      if (nEndPos != String::npos)
#pragma warning(suppress: 26472)
        nCurrentPos = static_cast<int>(nEndPos) + 1;
      else
        nCurrentPos = -1;
    }
  }

  _NODISCARD static String ChecksumField(_In_ StringArray& sFields)
  {
    //What will be the return value from this function
    String sChecksum;

    const auto nFields{sFields.size()};
    if (nFields > 1)
#pragma warning(suppress: 26446)
      sChecksum = sFields[nFields - 1];

    return sChecksum;
  }

#pragma warning(suppress: 26440)
  _NODISCARD static int Integer(_In_ StringArray& sFields, _In_ INT_PTR nField)
  {
#pragma warning(suppress: 26446)
    const String& sInteger{sFields[nField]};
    return _ttoi(sInteger.c_str());
  }

#pragma warning(suppress: 26440)
  _NODISCARD static int Integer(_In_ StringArray& sFields, _In_ INT_PTR nField, _Out_ bool& bValid)
  {
#pragma warning(suppress: 26446)
    const String& sInteger{sFields[nField]};
    if (sInteger.length())
    {
      bValid = true;
      return _ttoi(sInteger.c_str());
    }
    else
    {
      bValid = false;
      return 0;
    }
  }

  _NODISCARD static BOOLEAN Boolean(_In_ StringArray& sFields, _In_ INT_PTR nField)
  {
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    if (sField == _T("A"))
      return BOOLEAN::NMEA_TRUE;
#pragma warning(suppress: 26486 26489)
    else if (sField == _T("V"))
      return BOOLEAN::NMEA_FALSE;
    else
      return BOOLEAN::NMEA_UNKNOWN;
  }

  _NODISCARD static SELECTION_MODE SelectionMode(_In_ StringArray& sFields, _In_ INT_PTR nField)
  {
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    if (sField == _T("A"))
      return SELECTION_MODE::NMEA_SELECTION_MODE_AUTOMATIC;
#pragma warning(suppress: 26486 26489)
    else if (sField == _T("M"))
      return SELECTION_MODE::NMEA_SELECTION_MODE_MANUAL;
    else
      return SELECTION_MODE::NMEA_SELECTION_MODE_UNKNOWN;
  }

#pragma warning(suppress: 26440)
  _NODISCARD static double Double(_In_ StringArray& sFields, _In_ int nField, _Out_ bool& bValid)
  {
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    if (sField.length())
    {
      bValid = true;
      return _tstof(sField.c_str());
    }
    else
    {
      bValid = false;
      return 0;
    }
  }

  _NODISCARD static EASTWEST EastOrWest(_In_ StringArray& sFields, _In_ int nField)
  {
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    if (sField == _T("E"))
      return EASTWEST::NMEA_EAST;
#pragma warning(suppress: 26486 26489)
    else if (sField == _T("W"))
      return EASTWEST::NMEA_WEST;
    else
      return EASTWEST::NMEA_EW_UNKNOWN;
  }

  _NODISCARD static NORTHSOUTH NorthOrSouth(_In_ StringArray& sFields, _In_ int nField)
  {
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    if (sField == _T("N"))
      return NORTHSOUTH::NMEA_NORTH;
#pragma warning(suppress: 26486 26489)
    else if (sField == _T("S"))
      return NORTHSOUTH::NMEA_SOUTH;
    else
      return NORTHSOUTH::NMEA_NS_UNKNOWN;
  }

  _NODISCARD static TIME Time(_In_ StringArray& sFields, _In_ int nField)
  {
    TIME rVal{};
    rVal.nHour = -1;
    rVal.nMinute = -1;
    rVal.nSecond = -1;
    rVal.nCentiSeconds = -1;
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    const auto nLength{sField.length()};
    if (nLength >= 6)
    {
      rVal.nHour = _ttoi(sField.substr(0, 2).c_str());
#pragma warning(suppress: 26489)
      rVal.nMinute = _ttoi(sField.substr(2, 2).c_str());
#pragma warning(suppress: 26489)
      rVal.nSecond = _ttoi(sField.substr(4, 2).c_str());
      if (nLength >= 9)
#pragma warning(suppress: 26489)
        rVal.nCentiSeconds = _ttoi(sField.substr(7, 2).c_str());
      else
        rVal.nCentiSeconds = 0;
    }
    return rVal;
  }

  _NODISCARD static DATE Date(_In_ StringArray& sFields, _In_ int nField)
  {
    //What will be the return value from this function
    DATE rVal{};
    rVal.nYear = -1;
    rVal.nMonth = -1;
    rVal.nDay = -1;
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    if (sField.length() >= 6)
    {
      rVal.nDay = _ttoi(sField.substr(0, 2).c_str());
#pragma warning(suppress: 26489)
      rVal.nMonth = _ttoi(sField.substr(2, 2).c_str());
#pragma warning(suppress: 26489)
      rVal.nYear = _ttoi(sField.substr(4, 2).c_str());

      //intelligently adjust the year
      if (rVal.nYear < 80)
        rVal.nYear += 2000;
      else if (rVal.nYear >= 80)
        rVal.nYear += 1900;
    }
    return rVal;
  }

  _NODISCARD static BOOLEAN IsChecksumValid(_In_ const String& sSentence, _In_ StringArray& sFields)
  {
    BOOLEAN nBoolean{BOOLEAN::NMEA_UNKNOWN};

    //Compare the received checksum with the computed checksum
    const BYTE dwCalculatedCheckSum{ComputeChecksum(sSentence)};
    const DWORD dwReceivedCheckSum{HexValue(ChecksumField(sFields))};
    if (dwCalculatedCheckSum == dwReceivedCheckSum) 
      nBoolean = BOOLEAN::NMEA_TRUE;
    else
      nBoolean = BOOLEAN::NMEA_FALSE;
    return nBoolean;
  }

  _NODISCARD static LONGITUDE Longitude(_In_ StringArray& sFields, _In_ int nPositionFieldNumber, _In_ int nEastingFieldNumber)
  {
    LONGITUDE rVal{};
    rVal.Value = Angle(sFields, nPositionFieldNumber);
    rVal.Easting = EastOrWest(sFields, nEastingFieldNumber);
    rVal.bValid = rVal.Value.bValid && (rVal.Easting != EASTWEST::NMEA_EW_UNKNOWN);
    switch (rVal.Easting)
    {
      case EASTWEST::NMEA_EAST:
      {
        rVal.dValue = rVal.Value.dValue;
        break;
      }
      case EASTWEST::NMEA_WEST:
      {
        rVal.dValue = -rVal.Value.dValue;
        break;
      }
      default:
      {
        rVal.dValue = 0;
        break;
      }
    }
    return rVal;
  }

  _NODISCARD static LATITUDE Latitude(_In_ StringArray& sFields, _In_ int nPositionFieldNumber, _In_ int nNorthingFieldNumber)
  {
    LATITUDE rVal{};
    rVal.Value = Angle(sFields, nPositionFieldNumber);
    rVal.Northing = NorthOrSouth(sFields, nNorthingFieldNumber);
    rVal.bValid = rVal.Value.bValid && (rVal.Northing != NORTHSOUTH::NMEA_NS_UNKNOWN);
    switch (rVal.Northing)
    {
      case NORTHSOUTH::NMEA_NORTH:
      {
        rVal.dValue = rVal.Value.dValue;
        break;
      }
      case NORTHSOUTH::NMEA_SOUTH:
      {
        rVal.dValue = -rVal.Value.dValue;
        break;
      }
      default:
      {
        rVal.dValue = 0;
        break;
      }
    }
    return rVal;
  }

  _NODISCARD static ANGLE Angle(_In_ StringArray& sFields, _In_ int nField)
  {
    ANGLE rVal{};
    rVal.bValid = false;
#pragma warning(suppress: 26446)
    const String& sField{sFields[nField]};
    const auto nPointIndex{sField.find(_T('.'))};
    if (nPointIndex != String::npos)
    {
      rVal.bValid = true;
#pragma warning(suppress: 26489)
      String sUnits{sField.substr(0, nPointIndex)};
      const int nValue{_ttoi(sUnits.c_str())};
#pragma warning(suppress: 26472)
      rVal.wDegrees = static_cast<WORD>(nValue/100);
#pragma warning(suppress: 26472)
      rVal.wMinutes = static_cast<WORD>(nValue - rVal.wDegrees*100);
      String sFraction{sField.substr(nPointIndex, sField.length() - nPointIndex)};
      const double fValue{_tstof(sFraction.c_str())};
#pragma warning(suppress: 26467)
      rVal.wSeconds = static_cast<WORD>(fValue*60);
#pragma warning(suppress: 26467)
      rVal.wCentiSeconds = static_cast<WORD>((fValue*60 - rVal.wSeconds)*100.0);
      rVal.dValue = rVal.wDegrees + (rVal.wMinutes/60.0) + (fValue/60.0);
    }
    return rVal;
  }

#pragma warning(suppress: 26440)
  _NODISCARD static BYTE ComputeChecksum(_In_ const String& sSentence)
  {
    BYTE byChecksum{0};
    const auto nLength{sSentence.length()};
    String::size_type i{1}; //Skip over the $ at the begining of the sentence
#pragma warning(suppress: 26446)
    while ((i < nLength) && (sSentence[i] != _T('*')))
    {
#pragma warning(suppress: 26446 26472)
      byChecksum ^= static_cast<BYTE>(sSentence[i]);
      i++;
    }
    return byChecksum;
  }

#pragma warning(suppress: 26440)
  _NODISCARD static DWORD HexValue(_In_ const String& sHexString)
  {
    DWORD dwVal{0};
    int nMultiplier{1};
#pragma warning(suppress: 26472)
    const int nLen{static_cast<int>(sHexString.length())};
    for (int i{nLen-1}; i>=0; i--)
    {
#pragma warning(suppress: 26446)
      const TCHAR c{sHexString[i]};
#pragma warning(suppress: 26472)
      dwVal += (HexDigitToInt(static_cast<char>(c)) * nMultiplier);
      nMultiplier *= 16;
    }
    return dwVal;
  }

  _NODISCARD constexpr static int HexDigitToInt(_In_ char c) noexcept
  {
    int nValue{0};
    if ((c >= 'a') && (c <= 'f'))
      nValue =  c - 'a' + 10;
    else if ((c >= 'A') && (c <= 'F'))
      nValue =  c - 'A' + 10;
    else if ((c >= '0') && (c <= '9'))
      nValue = c - '0';
    return nValue;
  }
};


//A RMC sentence
class CRMCSentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 10)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_Time = Time(sFields, 1);
    m_Status = Boolean(sFields, 2);
#pragma warning(suppress: 26486 26489)
    m_Latitude = Latitude(sFields, 3, 4);
#pragma warning(suppress: 26486 26489)
    m_Longitude = Longitude(sFields, 5, 6);
#pragma warning(suppress: 26486 26489)
    m_dSpeedOverGround = Double(sFields, 7, m_bSpeedOverGroundValid);
#pragma warning(suppress: 26486 26489)
    m_dBearing = Double(sFields, 8, m_bBearingValid);
#pragma warning(suppress: 26486 26489)
    m_Date = Date(sFields, 9);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  TIME m_Time{}; //UTC Time
  BOOLEAN m_Status{BOOLEAN::NMEA_UNKNOWN}; //Status
  LATITUDE m_Latitude{}; //Latitude
  LONGITUDE m_Longitude{}; //Longitude
  double m_dSpeedOverGround{0}; //Speed over ground (knots)
  bool m_bSpeedOverGroundValid{false}; //Is m_dSpeedOverGround valid
  double m_dBearing{0}; //Bearing (0 - 360) degrees
  bool m_bBearingValid{false}; //Is m_dBearing valid
  DATE m_Date{}; //Date
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A GGA sentence
class CGGASentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 12)
      return false;

    //Then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_Time = Time(sFields, 1);
    m_Latitude = Latitude(sFields, 2, 3);
#pragma warning(suppress: 26486 26489)
    m_Longitude = Longitude(sFields, 4, 5);
#pragma warning(suppress: 26486 26489)
    m_Fix = static_cast<FIX>(Integer(sFields, 6));
#pragma warning(suppress: 26486 26489)
    m_nSatellites = Integer(sFields, 7);
#pragma warning(suppress: 26486 26489)
    m_dHDOP = Double(sFields, 8, m_bHDOPValid);
#pragma warning(suppress: 26486 26489)
    m_dAntennaAltitude = Double(sFields, 9, m_bAntennaAltitudeValid);
#pragma warning(suppress: 26486 26489)
    m_dGeoidalSeparation = Double(sFields, 11, m_bGeoidalSeparationValid);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  TIME m_Time{}; //UTC Time
  LATITUDE m_Latitude{}; //Latitude
  LONGITUDE m_Longitude{}; //Longitude
  FIX m_Fix{FIX::NMEA_FIX_NOT_AVAILABLE}; //GPS Quality Indicator
  int m_nSatellites{0}; //Number of satellites in view
  double m_dHDOP{0}; //Horizontal Dilution of precision (meters)
  bool m_bHDOPValid{false}; //Is the m_dHDOP value valid
  double m_dAntennaAltitude{0}; //Antenna altitude above/below mean-sea-level (geoid) (in meters)
  bool m_bAntennaAltitudeValid{false}; //Is the m_dAntennaAltitude value valid
  double m_dGeoidalSeparation{0}; //The difference between the WGS-84 earth ellipsoid and mean-sea-level (geoid), a negative value means mean-sea-level below ellipsoid
  bool m_bGeoidalSeparationValid{false}; //Is the m_dGeoidalSeparation value valid
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A specific element in a GSV NMEA group of sentences
class CGSVSentenceItem
{
public:
//Member variables
  int m_nSatellitePRN{-1}; //The satellite PRN number
  int m_nElevation{0}; //Elevation in degrees
  int m_nAzimuth{0}; //Azimuth in degrees
  int m_nSNR{0}; //The SNR in db
  bool m_bSNRValid{false}; //Is m_nSNR valid
};


//A grouping of GSV sentences
class CGSVSentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 3)
      return false;

    //Then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_nTotalSentences = Integer(sFields, 1);
    m_nOriginNumber = Integer(sFields, 2);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    //If this is the first sentence in the group then reset the "temp" array
    if (m_nOriginNumber == 1)
      m_tempSatellites.clear();

    //Parse all the remaining fields
#pragma warning(suppress: 26489)
    const auto nFields{sFields.size()};
    StringArray::size_type nCurrentField{4};
    while ((nCurrentField + 4) <= nFields)
    {
      CGSVSentenceItem satellite;
#pragma warning(suppress: 26486 26489)
      satellite.m_nSatellitePRN = Integer(sFields, nCurrentField);
#pragma warning(suppress: 26486 26489)
      satellite.m_nElevation = Integer(sFields, nCurrentField + 1);
#pragma warning(suppress: 26486 26489)
      satellite.m_nAzimuth = Integer(sFields, nCurrentField + 2);
#pragma warning(suppress: 26486 26489)
      satellite.m_nSNR = Integer(sFields, nCurrentField + 3, satellite.m_bSNRValid);

      //Add to the temp array
      m_tempSatellites.push_back(satellite);

      //Prepare for the next time around
      nCurrentField += 4;
    }

    //Only update the client facing array when we have all the sentences in the group
    if (m_nOriginNumber == m_nTotalSentences)
      m_Satellites = m_tempSatellites;

    return true;
  }

//Member variables
  int m_nTotalSentences{0}; //The total number of GSV sentences in this group
  int m_nOriginNumber{0}; //The origin number of this GSV sentence with this group
  std::vector<CGSVSentenceItem> m_tempSatellites; //The array of "satellites in view" (note this array is updated as each sentence is received)
  std::vector<CGSVSentenceItem> m_Satellites; //The array of "satellites in view" (note this array is only updated as when the last sentence of the group is received)
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A GLL sentence
class CGLLSentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 7)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_Latitude  = Latitude(sFields, 1, 2);
    m_Longitude = Longitude(sFields, 3, 4);
#pragma warning(suppress: 26486 26489)
    m_Time = Time(sFields, 5);
#pragma warning(suppress: 26486 26489)
    m_Status = Boolean(sFields, 6);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  LATITUDE m_Latitude{}; //Latitude
  LONGITUDE m_Longitude{}; //Longitude
  TIME m_Time{}; //UTC Time
  BOOLEAN m_Status{BOOLEAN::NMEA_UNKNOWN}; //Status
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A VTG sentence
class CVTGSentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 8)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_dTrueTrack = Double(sFields, 2, m_bTrueTrackValid);
    m_dMagneticTrack = Double(sFields, 3, m_bMagneticTrackValid);
#pragma warning(suppress: 26486 26489)
    m_dSpeedKnots = Double(sFields, 5, m_bSpeedKnotsValid);
#pragma warning(suppress: 26486 26489)
    m_dSpeedKMH = Double(sFields, 7, m_bSpeedKMHValid);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  double m_dTrueTrack{0}; //True Track
  bool m_bTrueTrackValid{false}; //Is m_dTrueTrack valid
  double m_dMagneticTrack{0}; //True Track
  bool m_bMagneticTrackValid{false}; //Is m_dMagneticTrack valid
  double m_dSpeedKnots{0}; //Speed knots
  bool m_bSpeedKnotsValid{false}; //Is m_dSpeedKnots valid 
  double m_dSpeedKMH{0}; //Speed kilometers Per Hour
  bool m_bSpeedKMHValid{false}; //Is m_dSpeedKMH valid
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A GSA sentence
class CGSASentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 3)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_SelectionMode  = SelectionMode(sFields, 1);
    m_Mode  = static_cast<MODE>(Integer(sFields, 2));
    m_Satellites.clear();
#pragma warning(suppress: 26489)
    const StringArray::size_type nMaxField{min(sFields.size() - 1, 14)};
    for (StringArray::size_type i{3}; i<=nMaxField; i++)
    {
#pragma warning(suppress: 26446 26489)
      const String& sField{sFields[i]};
      if (sField.length())
      {
        const int nSatellite{_ttoi(sField.c_str())};
        m_Satellites.push_back(nSatellite);
      }
    }

#pragma warning(suppress: 26486 26489)
    m_dPDOP = Double(sFields, 15, m_bPDOPValid);
#pragma warning(suppress: 26486 26489)
    m_dHDOP = Double(sFields, 16, m_bHDOPValid);
#pragma warning(suppress: 26486 26489)
    m_dVDOP = Double(sFields, 17, m_bVDOPValid);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  SELECTION_MODE m_SelectionMode{SELECTION_MODE::NMEA_SELECTION_MODE_UNKNOWN}; //Selection mode
  MODE m_Mode{MODE::NMEA_MODE_UNKNOWN}; //Mode
  std::vector<int> m_Satellites; //The ID of all the satellites used for fix
  double m_dPDOP{0}; //Dilution of precision
  bool m_bPDOPValid{false}; //Is m_dPDOP valid
  double m_dHDOP{0}; //Horizontal dilution of precision
  bool m_bHDOPValid{false}; //Is m_dHDOP valid
  double m_dVDOP{0}; //Vertical dilution of precision
  bool m_bVDOPValid{false}; //Is m_dVDOP valid
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A GRS sentence
class CGRSSentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 3)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_Time = Time(sFields, 1);
    m_Use  = static_cast<RESIDUAL_USE>(Integer(sFields, 2));
    m_Residuals.clear();
#pragma warning(suppress: 26489)
    const StringArray::size_type nMaxField{min(sFields.size() - 1, 14)};
    for (StringArray::size_type i{3}; i<=nMaxField; i++)
    {
#pragma warning(suppress: 26446 26489)
      const String& sField{sFields[i]};
      if (sField.length())
      {
        const int nResidual{_ttoi(sField.c_str())};
        m_Residuals.push_back(nResidual);
      }
    }
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  TIME m_Time{}; //UTC Time
  RESIDUAL_USE m_Use{RESIDUAL_USE::NMEA_RESIDUAL_USE_UNKNOWN}; //Residual use
  std::vector<int> m_Residuals; //The array of residual values in meters
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A GST sentence
class CGSTSentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 9)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_Time = Time(sFields, 1);
    m_dStandardDeviationRangeInputs = Double(sFields, 2, m_bStandardDeviationRangeInputsValid);
#pragma warning(suppress: 26486 26489)
    m_dStandardDeviationSemiMajorAxis = Double(sFields, 3, m_bStandardDeviationSemiMajorAxisValid);
#pragma warning(suppress: 26486 26489)
    m_dStandardDeviationSemiMinorAxis = Double(sFields, 4, m_bStandardDeviationSemiMinorAxisValid);
#pragma warning(suppress: 26486 26489)
    m_dSemiMajorAxisOrientation = Double(sFields, 5, m_bSemiMajorAxisOrientationValid);
#pragma warning(suppress: 26486 26489)
    m_dStandardDeviationLatitudeError = Double(sFields, 6, m_bStandardDeviationLatitudeErrorValid);
#pragma warning(suppress: 26486 26489)
    m_dStandardDeviationLongitudeError = Double(sFields, 7, m_bStandardDeviationLongitudeErrorValid);
#pragma warning(suppress: 26486 26489)
    m_dStandardDeviationAltitudeError = Double(sFields, 8, m_bStandardDeviationAltitudeErrorValid);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  TIME m_Time{}; //UTC Time
  double m_dStandardDeviationRangeInputs{0}; //Total RMS standard deviation of ranges inputs to the navigation solution
  bool m_bStandardDeviationRangeInputsValid{false}; //Is m_dStandardDeviationRangeInputs valid
  double m_dStandardDeviationSemiMajorAxis{0}; //Standard deviation (meters) of semi-major axis of error ellipse
  bool m_bStandardDeviationSemiMajorAxisValid{false}; //Is m_dStandardDeviationSemiMajorAxis valid
  double m_dStandardDeviationSemiMinorAxis{0}; //Standard deviation (meters) of semi-minor axis of error ellipse
  bool m_bStandardDeviationSemiMinorAxisValid{false}; //Is m_dStandardDeviationSemiMinorAxis valid
  double m_dSemiMajorAxisOrientation{0}; //Orientation of semi-major axis of error ellipse (true north degrees)
  bool m_bSemiMajorAxisOrientationValid{false}; //Is m_dSemiMajorAxisOrientation valid
  double m_dStandardDeviationLatitudeError{0}; //Standard deviation (meters) of latitude error
  bool m_bStandardDeviationLatitudeErrorValid{false}; //Is m_dStandardDeviationLatitudeError valid
  double m_dStandardDeviationLongitudeError{0}; //Standard deviation (meters) of longitude error
  bool m_bStandardDeviationLongitudeErrorValid{false}; //Is m_dStandardDeviationLongitudeError valid
  double m_dStandardDeviationAltitudeError{0}; //Standard deviation (meters) of altitude error
  bool m_bStandardDeviationAltitudeErrorValid{false}; //Is m_dStandardDeviationAltitudeError valid
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A GNS sentence
class CGNSSentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 13)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_Time = Time(sFields, 1);
#pragma warning(suppress: 26486 26489)
    m_Latitude = Latitude(sFields, 2, 3);
#pragma warning(suppress: 26486 26489)
    m_Longitude = Longitude(sFields, 4, 5);
#pragma warning(suppress: 26446 26489)
    m_sModeIndicator = sFields[6];
#pragma warning(suppress: 26486 26489)
    m_nVehicles = Integer(sFields, 7);
#pragma warning(suppress: 26486 26489)
    m_dHDOP = Double(sFields, 8, m_bHDOPValid);
#pragma warning(suppress: 26486 26489)
    m_dAntennaAltitude = Double(sFields, 9, m_bAntennaAltitudeValid);
#pragma warning(suppress: 26486 26489)
    m_dGeoidSeparation = Double(sFields, 10, m_bGeoidSeparationValid);
#pragma warning(suppress: 26446 26489)
    m_sAgeOfDifferentialCorrections = sFields[11];
#pragma warning(suppress: 26446 26489)
    m_sDGNSSRefStation = sFields[12];
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

  //Constructors / Destructors
  CGNSSentence() noexcept : m_Time{},
    m_Latitude{},
    m_Longitude{},
    m_nVehicles{ 0 },
    m_dHDOP{ 0 },
    m_bHDOPValid{ false },
    m_dAntennaAltitude{ 0 },
    m_bAntennaAltitudeValid{ false },
    m_dGeoidSeparation{ 0 },
    m_bGeoidSeparationValid{ false }
  {
  }


//Member variables
  TIME m_Time{}; //UTC Time
  LATITUDE m_Latitude{}; //Latitude
  LONGITUDE m_Longitude{}; //Longitude
  String m_sModeIndicator; //Mode indicator
  int m_nVehicles{0}; //Number of SVs used for Navigation
  double m_dHDOP{0}; //Horizontal Dilution of precision (meters)
  bool m_bHDOPValid{false}; //Is the m_dHDOP value valid
  double m_dAntennaAltitude{0}; //Antenna altitude above/below mean-sea-level (geoid) (in meters)
  bool m_bAntennaAltitudeValid{false}; //Is the m_dAntennaAltitude value valid
  double m_dGeoidSeparation{0}; //Geoid Separation = Alt (HAE) - Alt(MSL)
  bool m_bGeoidSeparationValid{false}; //Is the m_dAntennaAltitude value valid
  String m_sAgeOfDifferentialCorrections; //Age of Differential Corrections
  String m_sDGNSSRefStation; //ID of DGNSS Reference Station
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//A ZDA sentence
class CZDASentence : public CSentence
{
public:
//Methods
  bool Parse(_In_ bool bValidateChecksum, _In_ const String& sSentence, _In_ StringArray& sFields)
  {
    //First we check the checksum if required
    if (bValidateChecksum)
    {
      const BOOLEAN nCheck{IsChecksumValid(sSentence, sFields)};
      if (nCheck != BOOLEAN::NMEA_TRUE)
        return false;
    }

    //Validate the size of sFields
    if (sFields.size() < 7)
      return false;

    //then the data
#pragma warning(suppress: 26446)
    m_sMessageID = sFields[0];
    m_Time = Time(sFields, 1);
#pragma warning(suppress: 26486 26489)
    m_nDay = Integer(sFields, 2);
#pragma warning(suppress: 26486 26489)
    m_nMonth = Integer(sFields, 3);
#pragma warning(suppress: 26486 26489)
    m_nYear = Integer(sFields, 4);
#pragma warning(suppress: 26486 26489)
    m_nLocalZoneHours = Integer(sFields, 5);
#pragma warning(suppress: 26486 26489)
    m_nLocalZoneMinutes = Integer(sFields, 6);
#pragma warning(suppress: 26489)
    m_sChecksum = ChecksumField(sFields);

    return true;
  }

//Member variables
  TIME m_Time{}; //UTC Time
  int m_nDay{0}; //UTC Day
  int m_nMonth{0}; //UTC Month
  int m_nYear{0}; //UTC year
  int m_nLocalZoneHours{0}; //Local zone hours
  int m_nLocalZoneMinutes{0}; //Local zone minutes
  String m_sChecksum; //Checksum
  String m_sMessageID; //Can be used to distinguish between GPS, GNSS, GLONASS, Galileo and BeiDou variants
};


//Class which provides a C++ client side encapsulation of GPSCom2
class CGPSCOM2CLIENT_EXT_CLASS CClient : ATL::IDispEventSimpleImpl<1, CClient, &__uuidof(GPSCom2::_IGPSConnectionEvents)>
{
public:
//Connection point sink map
#pragma warning(suppress: 26440)
  BEGIN_SINK_MAP(CClient) //NOLINT(modernize-avoid-c-arrays)
    SINK_ENTRY_INFO(1, __uuidof(GPSCom2::_IGPSConnectionEvents), 1, &CClient::_OnSentence, &Event1Info)
#pragma warning(suppress: 26477 26485)
  END_SINK_MAP()

//Constructors / Destructors
  CClient() noexcept : m_bValidateChecksum{true}
  {
  }
  CClient(const CClient&) = delete;
  CClient(CClient&&) = delete;
  virtual ~CClient()
  {
#pragma warning(suppress: 26447)
    Close();
  }

//Methods
  CClient& operator=(const CClient&) = delete;
  CClient& operator=(CClient&&) = delete;

#pragma warning(suppress: 26812)
  HRESULT Open(_In_ long nPort, _In_ long nBaudRate = 4800, _In_ GPSCom2::GPS_PARITY parity = GPSCom2::GPS_PARITY_NONE, _In_ long DataBits = 8, _In_ GPSCom2::GPS_STOP_BITS StopBits = GPSCom2::GPS_1_STOP_BIT, _In_ const String& sInitString = _T(""))
  {
    //Create the port object
    ATL::CComPtr<GPSCom2::IGPSPort> Port;
    HRESULT hr{Port.CoCreateInstance(__uuidof(GPSCom2::GPSPort), nullptr, CLSCTX_LOCAL_SERVER)};
    if (FAILED(hr))
      return hr;

    //And from the port object, create the connection object
  #ifdef _UNICODE
    BSTR bstrInitString{SysAllocString(sInitString.c_str())};
  #else
    BSTR bstrInitString{SysAllocString(ATL::CA2W{sInitString.c_str()})};
  #endif //#ifdef _UNICODE
    if (bstrInitString == nullptr)
      return E_OUTOFMEMORY;

    hr = Port->CreateConnection(nPort, nBaudRate, parity, DataBits, StopBits, bstrInitString, &m_Connection);
    SysFreeString(bstrInitString);
    if (FAILED(hr))
      return hr;

    //Hook up the outbound interface to this object using IDispEventSimpleImpl
    hr = DispEventAdvise(m_Connection, &__uuidof(GPSCom2::_IGPSConnectionEvents));
    if (FAILED(hr))
      return hr;

    return S_OK;
  }

  HRESULT Open2(_In_ long nPort, _In_ const String& sInitString = _T(""))
  {
    //Create the port object
    ATL::CComPtr<GPSCom2::IGPSPort> Port;
    HRESULT hr{Port.CoCreateInstance(__uuidof(GPSCom2::GPSPort), nullptr, CLSCTX_LOCAL_SERVER)};
    if (FAILED(hr))
      return hr;

    //And from the port object, create the connection object
#ifdef _UNICODE
    BSTR bstrInitString{SysAllocString(sInitString.c_str())};
#else
    BSTR bstrInitString{SysAllocString(ATL::CA2W(sInitString.c_str()))};
#endif //#ifdef _UNICODE
    hr = Port->CreateConnection2(nPort, bstrInitString, &m_Connection);
    SysFreeString(bstrInitString);
    if (FAILED(hr))
      return hr;

    //Hook up the outbound interface to this object using IDispEventSimpleImpl
    hr = DispEventAdvise(m_Connection, &__uuidof(GPSCom2::_IGPSConnectionEvents));
    if (FAILED(hr))
      return hr;

    return S_OK;
  }

  void Close()
  {
    //Unhook the outbound interface from this object
    DispEventUnadvise(m_Connection, &__uuidof(GPSCom2::_IGPSConnectionEvents));

    //Release the connection object
    m_Connection.Release();
  }

#pragma warning(suppress: 26440)
  virtual void OnRMC(_In_ const String& /*sSentence*/, _In_ const CRMCSentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnGGA(_In_ const String& /*sSentence*/, _In_ const CGGASentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnGLL(_In_ const String& /*sSentence*/, _In_ const CGLLSentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnGSV(_In_ const String& /*sSentence*/, _In_ const CGSVSentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnVTG(_In_ const String& /*sSentence*/, _In_ const CVTGSentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnGSA(_In_ const String& /*sSentence*/, _In_ const CGSASentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnGRS(_In_ const String& /*sSentence*/, _In_ const CGRSSentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnGST(_In_ const String& /*sSentence*/, _In_ const CGSTSentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnGNS(_In_ const String& /*sSentence*/, _In_ const CGNSSentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnZDA(_In_ const String& /*sSentence*/, _In_ const CZDASentence& /*sentence*/)
  {
  }

#pragma warning(suppress: 26440)
  virtual void OnOtherSentence(_In_ const String& sSentence)
  {
    //For display purposes display any unparsed sentences to the output window
    ATLTRACE(_T("Unhandled sentence %s\n"), sSentence.c_str());
  }

  virtual void OnSentence(_In_ const String& sSentence)
  {
    //Break the sentence apart into its constituent fields
#pragma warning(suppress: 26489)
    StringArray sFields;
    CSentence::Tokenize(sSentence, sFields);

    if (sFields.size())
    {
      //determine which parser to call based on the first field value
#pragma warning(suppress: 26446)
      const String& sPrefix{sFields[0]};
#pragma warning(suppress: 26486 26489)
      if ((sPrefix == _T("$GPRMC")) || (sPrefix == _T("$GNRMC")) || (sPrefix == _T("$GLRMC")) || (sPrefix == _T("$GARMC")) || (sPrefix == _T("$GBRMC"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_RMC.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnRMC(sSentence, m_RMC);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPGGA")) || (sPrefix == _T("$GNGGA")) || (sPrefix == _T("$GLGGA")) || (sPrefix == _T("$GAGGA")) || (sPrefix == _T("$GBGGA"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_GGA.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnGGA(sSentence, m_GGA);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPGSV")) || (sPrefix == _T("$GNGSV")) || (sPrefix == _T("$GLGSV")) || (sPrefix == _T("$GAGSV")) || (sPrefix == _T("$GBGSV"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_GSV.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnGSV(sSentence, m_GSV);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPGLL")) || (sPrefix == _T("$GNGLL")) || (sPrefix == _T("$GLGLL")) || (sPrefix == _T("$GAGLL")) || (sPrefix == _T("$GBGLL"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_GLL.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnGLL(sSentence, m_GLL);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPVTG")) || (sPrefix == _T("$GNVTG")) || (sPrefix == _T("$GLVTG")) || (sPrefix == _T("$GAVTG")) || (sPrefix == _T("$GBVTG"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_VTG.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnVTG(sSentence, m_VTG);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPGSA")) || (sPrefix == _T("$GNGSA")) || (sPrefix == _T("$GLGSA")) || (sPrefix == _T("$GAGSA")) || (sPrefix == _T("$GBGSA"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_GSA.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnGSA(sSentence, m_GSA);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPGRS")) || (sPrefix == _T("$GNGRS")) || (sPrefix == _T("$GLGRS")) || (sPrefix == _T("$GAGRS")) || (sPrefix == _T("$GBGRS"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_GRS.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnGRS(sSentence, m_GRS);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPGST")) || (sPrefix == _T("$GNGST")) || (sPrefix == _T("$GLGST")) || (sPrefix == _T("$GAGST")) || (sPrefix == _T("$GBGST"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_GST.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnGST(sSentence, m_GST);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPGNS")) || (sPrefix == _T("$GNGNS")) || (sPrefix == _T("$GLGNS")) || (sPrefix == _T("$GAGNS")) || (sPrefix == _T("$GBGNS"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_GNS.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnGNS(sSentence, m_GNS);
      }
#pragma warning(suppress: 26486 26489)
      else if ((sPrefix == _T("$GPZDA")) || (sPrefix == _T("$GNZDA")) || (sPrefix == _T("$GLZDA")) || (sPrefix == _T("$GAZDA")) || (sPrefix == _T("$GBZDA"))) //Support GPS, GNSS, GLONASS, Galileo and BeiDou variants
      {
#pragma warning(suppress: 26489)
        if (m_ZDA.Parse(m_bValidateChecksum, sSentence, sFields))
#pragma warning(suppress: 26486 26489)
          OnZDA(sSentence, m_ZDA);
      }
      else
      {
        OnOtherSentence(sSentence);
      }
    }
  }

  void SetValidateChecksum(_In_ bool bValidateChecksum) noexcept { m_bValidateChecksum = bValidateChecksum; };
  bool GetValidateChecksum() const noexcept { return m_bValidateChecksum; };

//Member variables
  CRMCSentence m_RMC;
  CGGASentence m_GGA;
  CGSVSentence m_GSV;
  CGLLSentence m_GLL;
  CVTGSentence m_VTG;
  CGSASentence m_GSA;
  CGRSSentence m_GRS;
  CGSTSentence m_GST;
  CGNSSentence m_GNS;
  CZDASentence m_ZDA;

protected:
//Methods
#pragma warning(suppress: 26461)
  HRESULT __stdcall _OnSentence(_In_ BSTR bstrValue)
  {
    //Just delegate to the "OnSentence" method
  #ifdef _UNICODE
    OnSentence(String{bstrValue});
  #else
    OnSentence(String{ATL::CW2A{bstrValue}});
  #endif //#ifdef _UNICODE
    return S_OK;
  }

//Member variables
  ATL::CComPtr<GPSCom2::IGPSConnection> m_Connection;
  bool m_bValidateChecksum;
};

}; //namespace GPSCom2


#endif //__GPSCOM2CLIENT_H__
