/*
Module : GPSConnection.cpp
Purpose: Connection class for GPSCom2
Created: PJN / 27-04-2010

Copyright (c) 2010 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "GPSConnection.h"


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26434 26436 26440)
void CGPSConnection::FinalRelease()
{
  Close();
}

#pragma warning(suppress: 26433 26487)
STDMETHODIMP CGPSConnection::GetSettings(long* pnPort, long* pnDataBits, long* pnBaudRate, GPS_STOP_BITS* pStopBits, GPS_PARITY* pParity, BSTR* pbstrInitString) //NOLINT(clang-diagnostic-microsoft-exception-spec)
{
  //Validate our parameters
  if (pnPort == nullptr)
    return Error(_T("Port is a null pointer"), IID_IGPSConnection, E_POINTER);
  if (pnDataBits == nullptr)
    return Error(_T("DataBits is a null pointer"), IID_IGPSConnection, E_POINTER);
  if (pnBaudRate == nullptr)
    return Error(_T("BaudRate is a null pointer"), IID_IGPSConnection, E_POINTER);
  if (pStopBits == nullptr)
    return Error(_T("StopBits is a null pointer"), IID_IGPSConnection, E_POINTER);
  if (pParity == nullptr)
    return Error(_T("Parity is a null pointer"), IID_IGPSConnection, E_POINTER);
  if (pbstrInitString == nullptr)
    return Error(_T("InitString is a null pointer"), IID_IGPSConnection, E_POINTER);

  *pnPort = m_nPort;
  *pnDataBits = m_nDataBits;
  *pnBaudRate = m_nBaudRate;
  *pStopBits = m_StopBits;
  *pParity = m_Parity;
  return m_bstrInitString.CopyTo(pbstrInitString);
}

#pragma warning(suppress: 26461)
HRESULT CGPSConnection::Open(_In_ long nPort, _In_ BSTR bstrInitString)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_Port.operator HANDLE() == nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(m_StopEvent.operator HANDLE() == nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(m_DataEvent.operator HANDLE() == nullptr);

  //Create the events we will be using in the monitor thread
  if (!m_StopEvent.Create(nullptr, TRUE, FALSE, nullptr))
  {
    //Form the rich error information
    const DWORD dwError{GetLastError()};
    Close();
    return Error(_T("Failed to create stop event"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwError));
  }
  if (!m_DataEvent.Create(nullptr, FALSE, FALSE, nullptr))
  {
    //Form the rich error information
    const DWORD dwError{GetLastError()};
    Close();
    return Error(_T("Failed to create data event"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwError));
  }

  //Call CreateFile to open the comm port
  ATL::CString sPort;
  sPort.Format(_T("\\\\.\\COM%d"), nPort);
  HANDLE hComm{CreateFile(sPort, GENERIC_READ | GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, nullptr)};
  if (hComm == INVALID_HANDLE_VALUE)
  {
    //Form the rich error information
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to open specified port"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }
  m_Port.Attach(hComm);

  //Set the timeouts for the port
  COMMTIMEOUTS ct{};
  ct.ReadIntervalTimeout = MAXDWORD;
  ct.ReadTotalTimeoutConstant = 0;
  ct.ReadTotalTimeoutMultiplier = 0;
  ct.WriteTotalTimeoutMultiplier = 0;
  ct.WriteTotalTimeoutConstant = 0;
  if (!SetCommTimeouts(m_Port, &ct))
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to set port timeouts"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Setup the event mask for the port
  if (!SetCommMask(m_Port, EV_RXCHAR))
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to set port event mask"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Create the monitoring thread
#pragma warning(suppress: 26477)
  ATLASSERT(m_MonitorThread.operator HANDLE() == nullptr);
  unsigned threadAddress{0};
#pragma warning(suppress: 26490)
  auto hThread{reinterpret_cast<HANDLE>(_beginthreadex(nullptr, 0, _MonitorThread, this, CREATE_SUSPENDED, &threadAddress))};
  if (hThread == nullptr)
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to create worker thread"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }
#pragma warning(suppress: 26486)
  m_MonitorThread.Attach(hThread);

  //Hive away the init string before we resume the monitor thread
  m_bstrInitString = bstrInitString;

#pragma warning(suppress: 26472 26486)
  if (ResumeThread(hThread) == static_cast<DWORD>(-1))
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to resume worker thread"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Store away the member variables
  m_nDataBits = -1;
  m_nBaudRate = -1;
  m_StopBits = GPS_STOP_BITS_NONE;
  m_Parity = GPS_PARITY_NONE;
  m_nPort = nPort;

  //everything ok, return success indicator
  return S_OK;
}

#pragma warning(suppress: 26461)
HRESULT CGPSConnection::Open(_In_ long nPort, _In_ long nDataBits, _In_ long nBaudRate, _In_ GPS_STOP_BITS StopBits, _In_ GPS_PARITY Parity, _In_ BSTR bstrInitString)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_Port.operator HANDLE() == nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(m_StopEvent.operator HANDLE() == nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(m_DataEvent.operator HANDLE() == nullptr);

  //Create the events we will be using in the monitor thread
  if (!m_StopEvent.Create(nullptr, TRUE, FALSE, nullptr))
  {
    //Form the rich error information
    const DWORD dwError{GetLastError()};
    Close();
    return Error(_T("Failed to create stop event"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwError));
  }
  if (!m_DataEvent.Create(nullptr, FALSE, FALSE, nullptr))
  {
    //Form the rich error information
    const DWORD dwError{GetLastError()};
    Close();
    return Error(_T("Failed to create data event"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwError));
  }

  //Call CreateFile to open the comm port
  ATL::CString sPort;
  sPort.Format(_T("\\\\.\\COM%d"), nPort);
  HANDLE hComm{CreateFile(sPort, GENERIC_READ | GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, nullptr)};
  if (hComm == INVALID_HANDLE_VALUE)
  {
    //Form the rich error information
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to open specified port"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }
  m_Port.Attach(hComm);

  //Get the current state prior to changing it
  DCB dcb{};
  dcb.DCBlength = sizeof(DCB);
  if (!GetCommState(m_Port, &dcb))
  {
    //Form the rich error information
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to get port state"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Setup the baud rate
  dcb.BaudRate = nBaudRate;

  //Setup the stop bits
  switch (StopBits)
  {
    case GPS_1_STOP_BIT:
    {
      dcb.StopBits = ONESTOPBIT;
      break;
    }
    case GPS_1POINT5_STOP_BITS:
    {
      dcb.StopBits = ONE5STOPBITS;
      break;
    }
    case GPS_2_STOP_BITS:
    {
      dcb.StopBits = TWOSTOPBITS;
      break;
    }
    default:
    {
      Close();
      return Error(_T("StopBits parameter is invalid"), IID_IGPSConnection, E_INVALIDARG);
      break;
    }
  }

  //Setup the Parity
  switch (Parity)
  {
    case GPS_PARITY_EVEN:
    {
      dcb.Parity = EVENPARITY;
      break;
    }
    case GPS_PARITY_MARK:
    {
      dcb.Parity = MARKPARITY;
      break;
    }
    case GPS_PARITY_NONE:
    {
      dcb.Parity = NOPARITY;
      break;
    }
    case GPS_PARITY_ODD:
    {
      dcb.Parity = ODDPARITY;
      break;
    }
    case GPS_PARITY_SPACE:
    {
      dcb.Parity = SPACEPARITY;
      break;
    }
    default:
    {
      Close();
      return Error(_T("Parity parameter is invalid"), IID_IGPSConnection, E_INVALIDARG);
      break;
    }
  }

  //set the state of the port
  if (!SetCommState(m_Port, &dcb))
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to set port state"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Set the timeouts for the port
  COMMTIMEOUTS ct{};
  ct.ReadIntervalTimeout = MAXDWORD;
  ct.ReadTotalTimeoutConstant = 0;
  ct.ReadTotalTimeoutMultiplier = 0;
  ct.WriteTotalTimeoutMultiplier = 0;
  ct.WriteTotalTimeoutConstant = 0;
  if (!SetCommTimeouts(m_Port, &ct))
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to set port timeouts"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Setup the event mask for the port
  if (!SetCommMask(m_Port, EV_RXCHAR))
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to set port event mask"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Create the monitoring thread
#pragma warning(suppress: 26477)
  ATLASSERT(m_MonitorThread.operator HANDLE() == nullptr);
  unsigned threadAddress{0};
#pragma warning(suppress: 26490)
  auto hThread{reinterpret_cast<HANDLE>(_beginthreadex(nullptr, 0, _MonitorThread, this, CREATE_SUSPENDED, &threadAddress))};
  if (hThread == nullptr)
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to create worker thread"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }
#pragma warning(suppress: 26486)
  m_MonitorThread.Attach(hThread);

  //Hive away the init string before we resume the monitor thread
  m_bstrInitString = bstrInitString;

#pragma warning(suppress: 26472 26486)
  if (ResumeThread(hThread) == static_cast<DWORD>(-1))
  {
    const DWORD dwLastError{GetLastError()};
    Close();
    return Error(_T("Failed to resume worker thread"), IID_IGPSConnection, MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WINDOWS, dwLastError));
  }

  //Store away the member variables
  m_nDataBits = nDataBits;
  m_nBaudRate = nBaudRate;
  m_StopBits = StopBits;
  m_Parity = Parity;
  m_nPort = nPort;

  //everything ok, return success indicator
  return S_OK;
}

void CGPSConnection::Close() noexcept
{
  //Close down the monitoring thread
  if (m_MonitorThread != nullptr)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_StopEvent);

    //Request the worker thread to exit
    m_StopEvent.Set();

    //Synchronize with the exit of the monitor thread
    WaitForSingleObject(m_MonitorThread, INFINITE);
    m_MonitorThread.Close();
  }

  //Close down the comm port
  m_Port.Close();

  //Close the events
  m_StopEvent.Close();
  m_DataEvent.Close();
}

unsigned CGPSConnection::_MonitorThread(_In_ void* pParameter)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(pParameter != nullptr);

  //Initialize COM in this thread as we will be using connection points from it
  const HRESULT hr{CoInitialize(nullptr)};
  if (FAILED(hr))
    return 2;

  //Convert from the SDK world to the C++ world
  auto pThis{static_cast<CGPSConnection*>(pParameter)};
#pragma warning(suppress: 26477)
  ATLASSUME(pThis != nullptr);
  const unsigned nReturn{pThis->MonitorThread()}; //NOLINT(clang-analyzer-core.CallAndMessage)

  //Uninitialize COM in this thread
  CoUninitialize();

  return nReturn;
}

BOOL CGPSConnection::Write(_In_reads_bytes_opt_(dwSize) LPCVOID lpBuffer, _In_ DWORD dwSize, _In_ OVERLAPPED& overlapped) noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_Port != INVALID_HANDLE_VALUE);

  //Write out the data
  BOOL bSuccess{WriteFile(m_Port, lpBuffer, dwSize, nullptr, &overlapped)};
  if (!bSuccess && (GetLastError() == ERROR_IO_PENDING))
  {
    //Wait for the result if the operation has to be completed in overlapped mode
    DWORD dwBytesWritten{0};
    bSuccess = GetOverlappedResult(m_Port, &overlapped, &dwBytesWritten, TRUE);
  }

  return bSuccess;
}

void CGPSConnection::HandleRead(_In_z_ LPCSTR pszNewData, _Inout_ ATL::CStringA& sBuffer)
{
  //Append the newly read data to the buffer
  sBuffer += pszNewData;

  //See if we've got a sentence yet
  int nLocationDelimiter{sBuffer.Find('\n')};
  if (nLocationDelimiter == -1)
    nLocationDelimiter = sBuffer.Find('\r');

  while (nLocationDelimiter != -1)
  {
    //Yes, there is a line feed in the string
    ATL::CStringA sSentence{sBuffer.Left(nLocationDelimiter - 1)};

    //Lop off anything before the start of a NMEA sentence. This will handle the case
    //of any devices which mix NMEA sentences with other data. e.g. u-Blox devices
    const int nStartDelimiter{sSentence.Find("$G")};
    if (nStartDelimiter != -1)
      sSentence = sSentence.Mid(nStartDelimiter);

    //Now trim off the sentence from what came in from the port
    sBuffer = sBuffer.Right(sBuffer.GetLength() - (nLocationDelimiter + 1));
    nLocationDelimiter = sBuffer.Find('\n');
    if (nLocationDelimiter == -1)
      nLocationDelimiter = sBuffer.Find('\r');

    if (sSentence.GetLength() && (sSentence.Find("$G") == 0)) //Only fire the event for well-formed NMEA sentences
    {
      //Fire the event for the sentence
      BSTR bstrSentence{nullptr};
      ATLTRY(bstrSentence = sSentence.AllocSysString())
      if (bstrSentence != nullptr)
      {
        Fire_OnSentence(bstrSentence);
        SysFreeString(bstrSentence);
      }
    }
  }
}

void CGPSConnection::ReadData(_Inout_ ATL::CStringA& sBuffer, _Inout_ OVERLAPPED& overlapped)
{
  //Try to read something from the port in overlapped mode
  DWORD dwBytesRead{0};
  std::array<char, 4096> buffer{};
#pragma warning(suppress: 26472 26485)
  if (ReadFile(m_Port, buffer.data(), static_cast<DWORD>(buffer.size() - 1), &dwBytesRead, &overlapped) && dwBytesRead)
  {
    //Null terminate the received data
#pragma warning(suppress: 26446 26482)
    buffer[dwBytesRead] = '\0';

    //Process the received data
#pragma warning(suppress: 26485)
    HandleRead(buffer.data(), sBuffer);
  }
}

unsigned CGPSConnection::MonitorThread()
{
  //Setup the OVERLAPPED structure we will need
  OVERLAPPED overlapped{};
  overlapped.hEvent = m_DataEvent;

  //Send the init string out the port if we have one
  if (m_bstrInitString.Length())
  {
    ATL::CStringA szInit{m_bstrInitString};
#pragma warning(suppress: 26486)
    Write(szInit.GetString(), szInit.GetLength(), overlapped);
  }

  //Loop around waiting for something to do
  ATL::CStringA sBuffer;
  std::array<HANDLE, 2> handles{{ m_StopEvent, m_DataEvent }};
  bool bWantStop{false};
  while (!bWantStop)
  {
    DWORD dwEventMask{0};
#pragma warning(suppress: 26486)
    if (WaitCommEvent(m_Port, &dwEventMask, &overlapped))
      ReadData(sBuffer, overlapped);
    else if (GetLastError() == ERROR_IO_PENDING)
    {
      //If the read method failed asynchronously then wait on the 2 events
#pragma warning(suppress: 26485)
      const DWORD dwWait{WaitForMultipleObjects(2, handles.data(), FALSE, INFINITE)};
      if (dwWait == WAIT_FAILED)
        return 1;
      else
      {
        //Work out which event got signalled
        const DWORD dwSignaledHandle{dwWait - WAIT_OBJECT_0};
        if (dwSignaledHandle == 0)
          bWantStop = true;
        else
          ReadData(sBuffer, overlapped);
      }
    }
  }

  return 0;
}
